'use strict';

var assign = require('lodash/object/assign');
assign = 'default' in assign ? assign['default'] : assign;
var pick = require('lodash/object/pick');
pick = 'default' in pick ? pick['default'] : pick;
var sourceMap = require('source-map');
var _ = require('lodash');
var detectIndent = require('detect-indent');
detectIndent = 'default' in detectIndent ? detectIndent['default'] : detectIndent;
var matchesSelector = require('dom-matches');
matchesSelector = 'default' in matchesSelector ? matchesSelector['default'] : matchesSelector;
var merge = require('plain-merge');
merge = 'default' in merge ? merge['default'] : merge;

var babelHelpers = {};
babelHelpers.typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) {
  return typeof obj;
} : function (obj) {
  return obj && typeof Symbol === "function" && obj.constructor === Symbol ? "symbol" : typeof obj;
};

babelHelpers.classCallCheck = function (instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
};

babelHelpers.createClass = function () {
  function defineProperties(target, props) {
    for (var i = 0; i < props.length; i++) {
      var descriptor = props[i];
      descriptor.enumerable = descriptor.enumerable || false;
      descriptor.configurable = true;
      if ("value" in descriptor) descriptor.writable = true;
      Object.defineProperty(target, descriptor.key, descriptor);
    }
  }

  return function (Constructor, protoProps, staticProps) {
    if (protoProps) defineProperties(Constructor.prototype, protoProps);
    if (staticProps) defineProperties(Constructor, staticProps);
    return Constructor;
  };
}();

babelHelpers.inherits = function (subClass, superClass) {
  if (typeof superClass !== "function" && superClass !== null) {
    throw new TypeError("Super expression must either be null or a function, not " + typeof superClass);
  }

  subClass.prototype = Object.create(superClass && superClass.prototype, {
    constructor: {
      value: subClass,
      enumerable: false,
      writable: true,
      configurable: true
    }
  });
  if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass;
};

babelHelpers.possibleConstructorReturn = function (self, call) {
  if (!self) {
    throw new ReferenceError("this hasn't been initialised - super() hasn't been called");
  }

  return call && (typeof call === "object" || typeof call === "function") ? call : self;
};

babelHelpers;
var ASTNode = function () {
	function ASTNode(line, col, props) {
		babelHelpers.classCallCheck(this, ASTNode);

		this._line = line;
		this._column = col;
		assign(this, props);
	}

	babelHelpers.createClass(ASTNode, [{
		key: "_sn",
		value: function _sn(file, chunk) {
			return new sourceMap.SourceNode(this._line, this._column, file, chunk);
		}
	}, {
		key: "start",
		value: function start(data) {
			this._writer = {
				chunks: [],
				data: data
			};

			return this;
		}
	}, {
		key: "_normalize_indent",
		value: function _normalize_indent() {
			var d = this._writer.data;
			if (typeof d.indent === "string") d.tabchar = d.indent;
			if (d.indent !== false && typeof d.indent !== "number") d.indent = 0;
		}
	}, {
		key: "indent",
		value: function indent() {
			this._normalize_indent();
			var d = this._writer.data;
			if (typeof d.indent === "number") d.indent++;
			return this;
		}
	}, {
		key: "outdent",
		value: function outdent() {
			this._normalize_indent();
			var d = this._writer.data;
			if (typeof d.indent === "number" && d.indent > 0) d.indent--;
			return this;
		}
	}, {
		key: "tabs",
		value: function tabs() {
			this._normalize_indent();

			var res = "";
			var d = this._writer.data;
			var tabchar = d.tabchar;
			if (tabchar == null) tabchar = "  ";

			if (typeof d.indent === "number") {
				for (var i = 0; i < d.indent; i++) {
					res += tabchar;
				}
			}

			return res;
		}
	}, {
		key: "write",
		value: function write(chunk) {
			this.push([].concat(this.tabs(), chunk, "\n"));
			return this;
		}
	}, {
		key: "push",
		value: function push(chunk) {
			this._writer.chunks.push(chunk);
			return this;
		}
	}, {
		key: "end",
		value: function end(chunk) {
			if (chunk) this.write(chunk);
			var w = this._writer;
			delete this._writer;
			return this._sn(w.data.originalFilename, w.chunks);
		}
	}, {
		key: "compile",
		value: function compile() {
			throw new Error("Not implemented.");
			// return this._sn(data.originalFilename);
		}
	}, {
		key: "toJSON",
		value: function toJSON() {
			var _this = this;

			var out = {};
			Object.keys(this).forEach(function (k) {
				if (k[0] === "_") return;
				out[k] = _this[k];
			});
			assign(out, pick(this, "_line", "_column"));
		}
	}]);
	return ASTNode;
}();

var Comment = function (_Node) {
	babelHelpers.inherits(Comment, _Node);

	function Comment() {
		babelHelpers.classCallCheck(this, Comment);
		return babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(Comment).apply(this, arguments));
	}

	babelHelpers.createClass(Comment, [{
		key: "compile",
		value: function compile(data) {
			this.start(data);
			return this.end();
		}
	}]);
	return Comment;
}(ASTNode);

var Literal = function (_Node) {
	babelHelpers.inherits(Literal, _Node);

	function Literal() {
		babelHelpers.classCallCheck(this, Literal);
		return babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(Literal).apply(this, arguments));
	}

	babelHelpers.createClass(Literal, [{
		key: "compile",
		value: function compile(data) {
			this.start(data);
			this.push(JSON.stringify(this.value));
			return this.end();
		}
	}, {
		key: "reactive",
		get: function get() {
			return false;
		}
	}]);
	return Literal;
}(ASTNode);

function header(data, h) {
	if (_.isArray(data.headers) && !_.contains(data.headers, h)) {
		data.headers.push(h);
	}
}

function hash(str) {
	var h = 0,
	    i,
	    chr,
	    len;

	if (str.length === 0) return h;

	for (i = 0, len = str.length; i < len; i++) {
		chr = str.charCodeAt(i);
		h = (h << 5) - h + chr;
		h |= 0; // Convert to 32bit integer
	}

	return h;
}

function compileGroup(nodes, data) {
	return nodes.map(function (n, i) {
		return n.compile(addKey(data, i.toString()));
	});
}

function getKey(data) {
	return (data.key || []).reduce(function (memo, v) {
		if (_.isString(v) && _.isString(_.last(memo))) {
			memo.push(memo.pop() + "-" + v);
		} else {
			memo.push(v);
		}

		return memo;
	}, []).map(function (v, i, l) {
		if ((typeof v === "undefined" ? "undefined" : babelHelpers.typeof(v)) === "object") return v.value;
		if (i !== 0) v = "-" + v;
		if (i !== l.length - 1) v += "-";
		return JSON.stringify(v);
	}).join(" + ");
}

function addKey(data, key) {
	var base = data.key || [];
	return _.extend({}, data, {
		key: [].concat(base, key)
	});
}

function resetKey(data) {
	return _.extend({}, data, { key: [] });
}

function contextHeader(data, h) {
	if (_.isArray(data.contextHeaders) && !_.contains(data.contextHeaders, h)) {
		data.contextHeaders.push(h);
	}
}

function resetContextHeader(data) {
	return _.extend({}, data, { contextHeaders: [] });
}

var Expression = function (_Node) {
	babelHelpers.inherits(Expression, _Node);

	function Expression() {
		babelHelpers.classCallCheck(this, Expression);
		return babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(Expression).apply(this, arguments));
	}

	babelHelpers.createClass(Expression, [{
		key: "compile",
		value: function compile(data) {
			this.start(data);
			var exp = Expression.render(this.tree);
			var v = ("EXP_" + hash(exp)).replace("-", "_");

			if (Expression.isStatic(this.tree)) {
				if (data.asFn) {
					header(data, "var " + v + " = function(){return " + exp + ";};\n");
					this.push("" + v);
				} else {
					this.push(exp);
				}
			} else {
				header(data, "var " + v + " = function(c){return " + exp + ";};\n");

				if (data.asFn) {
					var cv = "B" + v;
					contextHeader(data, "var " + cv + " = function(){return " + v + "(ctx);};");
					this.push(cv);
				} else {
					this.push(v + "(ctx)");
				}
			}

			return this.end();
		}
	}, {
		key: "reactive",
		get: function get() {
			return true;
		}
	}], [{
		key: "isStatic",
		value: function isStatic(tree) {
			var _isStatic = Expression.isStatic;

			switch (tree.type) {
				case "Compound":
					return tree.body.every(_isStatic);
				case "Identifier":
					return false;
				case "ThisExpression":
					return false;
				case "MemberExpression":
					{
						if (!_isStatic(tree.object)) return false;
						var prop = tree.property;
						if (!tree.computed && prop.type === "Identifier") return true;
						return _isStatic(prop);
					}
				case "Literal":
					return true;
				case "CallExpression":
					{
						return _isStatic(tree.callee) && tree.arguments.every(_isStatic);
					}
				case "UnaryExpression":
					return _isStatic(tree.argument);
				case "LogicalExpression":
				case "BinaryExpression":
					return _isStatic(tree.left) && _isStatic(tree.right);
				case "ArrayExpression":
					return tree.elements.every(_isStatic);
				case "ConditionalExpression":
					{
						return [tree.test, tree.consequent, tree.alternate].every(_isStatic);
					}
				default:
					return false;
			}
		}
	}, {
		key: "render",
		value: function render(tree) {
			var _render = Expression.render;

			switch (tree.type) {
				case "Compound":
					{
						// expressions seperated by comma or semicolon
						return "(" + tree.body.map(_render).join(",") + ")";
					}

				case "Identifier":
					// plain varable
					return "c.lookup(" + JSON.stringify(tree.name) + ")";

				case "ThisExpression":
					// plain varable
					return "c.lookup()";

				case "MemberExpression":
					{
						// object property
						var out = _render(tree.object);
						var prop = tree.property;
						if (!tree.computed && prop.type === "Identifier") out += "." + prop.name;else out += "[" + _render(prop) + "]";
						return out;
					}

				case "Literal":
					// string, number, boolean
					return JSON.stringify(tree.value);

				case "CallExpression":
					{
						// function call
						var out = _render(tree.callee);
						if (tree.callee.type === "MemberExpression") out += "(";else out += ".call(c.lookup()" + (tree.arguments.length ? ", " : "");
						out += tree.arguments.map(_render).join(", ") + ")";
						return out;
					}

				case "UnaryExpression":
					{
						// single operators: ! - ++ --
						var out = _render(tree.argument);
						if (tree.prefix) out = tree.operator + out;else out += tree.operator;
						return out;
					}

				case "LogicalExpression":
				case "BinaryExpression":
					{
						// math
						return "(" + _render(tree.left) + tree.operator + _render(tree.right) + ")";
					}

				case "ArrayExpression":
					{
						// arrays
						return "[" + tree.elements.map(_render).join(",") + "]";
					}

				case "ConditionalExpression":
					{
						// if ? then : else
						return "(" + _render(tree.test) + "?" + _render(tree.consequent) + ":" + _render(tree.alternate) + ")";
					}

				default:
					throw new Error("Not supported.");
			}
		}
	}]);
	return Expression;
}(ASTNode);

var Interpolator = function (_Node) {
	babelHelpers.inherits(Interpolator, _Node);

	function Interpolator() {
		babelHelpers.classCallCheck(this, Interpolator);
		return babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(Interpolator).apply(this, arguments));
	}

	babelHelpers.createClass(Interpolator, [{
		key: "compile",
		value: function compile(data) {
			this.start(data);
			this.write("idom.autotext(" + this.expression.compile(_.assign({ asFn: true }, data)) + ");");
			return this.end();
		}
	}, {
		key: "reactive",

		// while technically reactive, they contain their own autorun statement
		get: function get() {
			return false;
		}
	}]);
	return Interpolator;
}(ASTNode);

var Render = function (_Node) {
	babelHelpers.inherits(Render, _Node);

	function Render() {
		babelHelpers.classCallCheck(this, Render);
		return babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(Render).apply(this, arguments));
	}

	babelHelpers.createClass(Render, [{
		key: "compile",
		value: function compile(data) {
			this.start(data);
			var exp = this.expression.compile(data);
			this.write(["Temple.render(", exp, ", ctx, " + getKey(data) + ");"]);
			return this.end();
		}
	}, {
		key: "reactive",
		get: function get() {
			return true;
		}
	}]);
	return Render;
}(ASTNode);

var Set = function (_Node) {
	babelHelpers.inherits(Set, _Node);

	function Set() {
		babelHelpers.classCallCheck(this, Set);
		return babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(Set).apply(this, arguments));
	}

	babelHelpers.createClass(Set, [{
		key: "compile",
		value: function compile(data) {
			this.start(data);

			var exp = this.expression.compile(data);
			var safevar = JSON.stringify(this.variable);
			if (this.attribute) this.push("(");
			this.push(["ctx.set(" + safevar + ", ", exp, ")"]);
			if (!this.attribute) this.write(";");else this.push(", \"\")");

			return this.end();
		}
	}, {
		key: "reactive",
		get: function get() {
			return true;
		}
	}]);
	return Set;
}(ASTNode);

var With = function (_Node) {
	babelHelpers.inherits(With, _Node);

	function With() {
		babelHelpers.classCallCheck(this, With);
		return babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(With).apply(this, arguments));
	}

	babelHelpers.createClass(With, [{
		key: "compile",
		value: function compile(data) {
			this.start(data);

			var exp = this.expression.compile(data);

			if (!this.attribute) this.push(this.tabs());
			this.push(["Temple.With(", exp, ", ctx, function(ctx) {"]).indent();
			if (!this.attribute) this.push("\n");
			data = resetContextHeader(data);
			var c = this._sn(data.originalFilename, compileGroup(this.children, data));
			data.contextHeaders.forEach(this.write, this);
			if (this.attribute) c = [" return Temple.utils.joinValues(", c.join(","), "); "];
			this.push(c);
			this.outdent().push("})");

			if (!this.attribute) this.write(";");

			return this.end();
		}
	}, {
		key: "reactive",
		get: function get() {
			return true;
		}
	}]);
	return With;
}(ASTNode);

var Each = function (_Node) {
	babelHelpers.inherits(Each, _Node);

	function Each() {
		babelHelpers.classCallCheck(this, Each);
		return babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(Each).apply(this, arguments));
	}

	babelHelpers.createClass(Each, [{
		key: "compile",
		value: function compile(data) {
			this.start(data);

			var exp = this.expression.compile(data);
			var indexvar = _.uniqueId("index");
			var vars = JSON.stringify(this.variables || []);

			if (!this.attribute) this.push(this.tabs());
			this.push(["Temple.Each(", exp, ", ", vars, ", ctx, function(ctx, " + indexvar + ") {"]).indent();
			if (!this.attribute) this.push("\n");
			data = resetContextHeader(data);
			data = addKey(data, { value: indexvar });
			var c = this._sn(data.originalFilename, compileGroup(this.children, data));
			data.contextHeaders.forEach(this.write, this);
			if (this.attribute) c = [" return Temple.utils.joinValues(", c.join(","), "); "];
			this.push(c);
			this.outdent().push("})");
			if (!this.attribute) this.write(";");else this.push(".join(\"\")");

			return this.end();
		}
	}, {
		key: "reactive",
		get: function get() {
			return true;
		}
	}]);
	return Each;
}(ASTNode);

var Branch = function (_Node) {
	babelHelpers.inherits(Branch, _Node);

	function Branch() {
		babelHelpers.classCallCheck(this, Branch);
		return babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(Branch).apply(this, arguments));
	}

	babelHelpers.createClass(Branch, [{
		key: "compile",
		value: function compile(data) {
			this.start(data);
			this.push(compileGroup(this.children, data));
			return this.end();
		}
	}]);
	return Branch;
}(ASTNode);

var If = function (_Node) {
	babelHelpers.inherits(If, _Node);

	function If() {
		babelHelpers.classCallCheck(this, If);
		return babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(If).apply(this, arguments));
	}

	babelHelpers.createClass(If, [{
		key: "compile",
		value: function compile(data) {
			var _this2 = this;

			this.start(data);

			if (this.attribute) {
				this.push("(");
			}

			this.children.forEach(function (c, i, l) {
				if (_this2.attribute) {
					if (c.expression) _this2.push([c.expression.compile(data), " ? "]);
					_this2.push(["Temple.utils.joinValues(", c.compile(addKey(data, i.toString())).join(","), ")"]);
					if (i + 1 === l.length && c.expression) _this2.push(" : ''");
				} else {
					var line = [];
					line.push(!i ? "if" : !c.expression ? "else" : "else if", " ");
					if (c.expression) line.push(["(", c.expression.compile(data), ") "]);
					line.push("{");
					_this2.write(line).indent();
					_this2.push(c.compile(addKey(data, i.toString())));
					_this2.outdent().write("}");
				}
			});

			if (this.attribute) {
				this.push(")");
			}

			return this.end();
		}
	}, {
		key: "reactive",
		get: function get() {
			return true;
		}
	}]);
	return If;
}(ASTNode);

var Attribute = function (_Node) {
	babelHelpers.inherits(Attribute, _Node);

	function Attribute() {
		babelHelpers.classCallCheck(this, Attribute);
		return babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(Attribute).apply(this, arguments));
	}

	babelHelpers.createClass(Attribute, [{
		key: "compile",
		value: function compile(data) {
			this.start(data);

			var value = undefined;

			if (this.children) {
				var str = undefined,
				    fun = undefined;

				if (this.children instanceof Expression) {
					str = this.children.compile(data);
					fun = true;
				} else {
					var len = this.children.length;
					data = resetContextHeader(data);
					str = _.invoke(this.children, "compile", data);
					str = !len ? null : this._sn(data.originalFilename, str).join(",");
					if (len > 1) str = ["Temple.utils.joinValues(", str, ")"];
					fun = len && this.children.some(function (c) {
						return c.reactive;
					});
					if (fun) str = ["[", str, "]"];
				}

				if (fun) {
					str = ["function(ctx){", data.contextHeaders.join("\n"), "return ", str, ";}"];
					value = _.uniqueId("ATTR_");
					header(data, ["var ", value, "=", str, ";\n"]);
				} else {
					value = str;
				}
			}

			this.write(["decorators.render(ctx, " + JSON.stringify(this.name), value != null ? [", ", value] : "", ");"]);

			return this.end();
		}
	}, {
		key: "reactive",
		get: function get() {
			return false;
		}
	}]);
	return Attribute;
}(ASTNode);

// like jQuery's empty(), removes all children
function emptyNode(node) {
	while (node.lastChild) {
		node.removeChild(node.lastChild);
	}return node;
}

var entity = /&(?:#x[a-f0-9]+|#[0-9]+|[a-z0-9]+);?/ig;

function containsEntities(str) {
	return entity.test(str);
}

// converts html entities to unicode
var decodeEntities = function () {
	if (typeof document === "undefined") return;

	// this prevents any overhead from creating the object each time
	var element = document.createElement('div');

	return function decodeHTMLEntities(str) {
		str = str.replace(entity, function (m) {
			element.innerHTML = m;
			return element.textContent;
		});

		emptyNode(element);

		return str;
	};
}();

var Text = function (_Node) {
	babelHelpers.inherits(Text, _Node);

	function Text() {
		babelHelpers.classCallCheck(this, Text);
		return babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(Text).apply(this, arguments));
	}

	babelHelpers.createClass(Text, [{
		key: "compile",
		value: function compile(data) {
			this.start(data);

			var value = this.value.replace(/\s+/g, " ");
			var safe = JSON.stringify(value);

			if (containsEntities(value)) {
				safe = "Temple.utils.decodeEntities(" + safe + ")";
			}

			this.write("idom.text(" + safe + ");");
			return this.end();
		}
	}]);
	return Text;
}(ASTNode);

var Element = function (_Node) {
	babelHelpers.inherits(Element, _Node);

	function Element() {
		babelHelpers.classCallCheck(this, Element);
		return babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(Element).apply(this, arguments));
	}

	babelHelpers.createClass(Element, [{
		key: "compile",
		value: function compile(data) {
			var _this2 = this;

			this.start(data);

			var tagName = JSON.stringify(this.tagname);
			var key = getKey(data);
			var childs = this.children;
			var attrs = this.attributes;

			var body = function body(newctx) {
				if (newctx) {
					data = resetContextHeader(data);
				}

				var c = [].concat(_.invoke(attrs, "compile", data), compileGroup(childs, resetKey(data)));

				if (newctx) {
					data.contextHeaders.forEach(_this2.write, _this2);
				}

				_this2.push(c);
			};

			if (childs.length || attrs.length) {
				if (!this.reactive) {
					this.write("idom.elementOpen(" + tagName + (key ? ", " + key : "") + ");");
					body();
					this.write("idom.elementClose(" + tagName + ");");
				} else {
					this.write("Temple.Element(" + tagName + ", " + (key || "null") + ", ctx, function(ctx) {").indent();
					body(true);
					this.outdent().write("});");
				}
			} else {
				this.write("idom.elementVoid(" + tagName + (key ? ", " + key : "") + ");");
			}

			return this.end();
		}
	}, {
		key: "reactive",
		get: function get() {
			return this.children.some(function (c) {
				return c.reactive;
			});
		}
	}]);
	return Element;
}(ASTNode);

var Script = function (_Node) {
	babelHelpers.inherits(Script, _Node);

	function Script() {
		babelHelpers.classCallCheck(this, Script);
		return babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(Script).apply(this, arguments));
	}

	babelHelpers.createClass(Script, [{
		key: "compile",
		value: function compile(data) {
			var _this2 = this;

			this.start(data);
			this._normalize_indent();

			var line = this._line || 1;
			var column = this._column || 1;
			var indent = detectIndent(this.value);

			this.value.split(/\r?\n/g).forEach(function (l, i) {
				var lindent = detectIndent(l);
				l = l.replace(/^\s+/, "");
				var len = Math.round(lindent.amount / indent.amount);
				var tabs = _this2.tabs();
				var tabchar = data.tabchar;
				if (tabchar == null) tabchar = "  ";
				for (var j = 0; j < len; j++) {
					tabs += tabchar;
				}_this2.push(new sourceMap.SourceNode(i + line, i === 0 ? column : 1, data.originalFilename, tabs + l + "\n"));
			});

			return this.end();
		}
	}, {
		key: "reactive",
		get: function get() {
			return true;
		}
	}]);
	return Script;
}(ASTNode);

var Template = function (_Node) {
	babelHelpers.inherits(Template, _Node);

	function Template() {
		babelHelpers.classCallCheck(this, Template);
		return babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(Template).apply(this, arguments));
	}

	babelHelpers.createClass(Template, [{
		key: "compile",
		value: function compile(data) {
			var _this2 = this;

			this.start(data);
			var safename = JSON.stringify(this.name);
			var safetype = this.type ? JSON.stringify(this.type) : null;

			this.write("Template[" + safename + "] = Temple.create(" + safename + ", " + (safetype ? safetype + ", " : "") + "function(ctx, key) {").indent();
			data = resetContextHeader(data);
			data = addKey(data, { value: "key" });
			var c = compileGroup(this.children, data);
			data.contextHeaders.forEach(this.write, this);
			this.push(c);
			this.outdent().write("});\n");

			if (this.plugins && this.plugins.length) {
				this.plugins.forEach(function (p) {
					_this2.write("Template[" + safename + "].use(" + JSON.stringify(p) + ");");
				});
				this.write("");
			}

			return this.end();
		}
	}]);
	return Template;
}(ASTNode);

var Root = function (_Node) {
	babelHelpers.inherits(Root, _Node);

	function Root() {
		babelHelpers.classCallCheck(this, Root);
		return babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(Root).apply(this, arguments));
	}

	babelHelpers.createClass(Root, [{
		key: "compile",
		value: function compile(data) {
			var oheads = data.headers;
			data.headers = [];

			this.start(data);
			header(data, "var Template = {};\n");
			header(data, "var idom = Temple.idom;\n");
			header(data, "var decorators = Temple.decorators;\n");
			this.push(_.invoke(this.children, "compile", data));

			if (data.exports === "es6") {
				this.write("export default Template;");
			} else if (data.exports === "cjs") {
				this.write("module.exports = Template;");
			} else if (data.exports === "none") {
				// print nothing
			} else {
					this.write("return Template;");
				}

			var output = this.end();
			if (data.headers.length) output.prepend("\n").prepend(data.headers);
			data.headers = oheads;

			return output;
		}
	}]);
	return Root;
}(ASTNode);

exports.Node = ASTNode;
exports.Root = Root;
exports.Template = Template;
exports.Script = Script;
exports.Element = Element;
exports.Text = Text;
exports.Attribute = Attribute;
exports.If = If;
exports.Branch = Branch;
exports.Each = Each;
exports.With = With;
exports.Set = Set;
exports.Render = Render;
exports.Interpolator = Interpolator;
exports.Expression = Expression;
exports.Literal = Literal;
exports.Comment = Comment;