'use strict';

var Trackr = require('trackr');
Trackr = 'default' in Trackr ? Trackr['default'] : Trackr;
var trackrObjects = require('trackr-objects');
var _ = require('lodash');
var matchesSelector = require('dom-matches');
matchesSelector = 'default' in matchesSelector ? matchesSelector['default'] : matchesSelector;
var merge = require('plain-merge');
merge = 'default' in merge ? merge['default'] : merge;
var events = require('events');
var assignProps = require('assign-props');
assignProps = 'default' in assignProps ? assignProps['default'] : assignProps;
var subclass = require('backbone-extend-standalone');
subclass = 'default' in subclass ? subclass['default'] : subclass;
var sourceMap = require('source-map');
var raf = require('raf');
raf = 'default' in raf ? raf['default'] : raf;

var babelHelpers = {};
babelHelpers.typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) {
  return typeof obj;
} : function (obj) {
  return obj && typeof Symbol === "function" && obj.constructor === Symbol ? "symbol" : typeof obj;
};

babelHelpers.classCallCheck = function (instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
};

babelHelpers.createClass = function () {
  function defineProperties(target, props) {
    for (var i = 0; i < props.length; i++) {
      var descriptor = props[i];
      descriptor.enumerable = descriptor.enumerable || false;
      descriptor.configurable = true;
      if ("value" in descriptor) descriptor.writable = true;
      Object.defineProperty(target, descriptor.key, descriptor);
    }
  }

  return function (Constructor, protoProps, staticProps) {
    if (protoProps) defineProperties(Constructor.prototype, protoProps);
    if (staticProps) defineProperties(Constructor, staticProps);
    return Constructor;
  };
}();

babelHelpers;
/**
 * Copyright 2015 The Incremental DOM Authors. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS-IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


/**
 * A cached reference to the hasOwnProperty function.
 */
var hasOwnProperty = Object.prototype.hasOwnProperty;


/**
 * A cached reference to the create function.
 */
var create$1 = Object.create;


/**
 * Used to prevent property collisions between our "map" and its prototype.
 * @param {!Object<string, *>} map The map to check.
 * @param {string} property The property to check.
 * @return {boolean} Whether map has property.
 */
var has = function(map, property) {
  return hasOwnProperty.call(map, property);
};


/**
 * Creates an map object without a prototype.
 * @return {!Object}
 */
var createMap = function() {
  return create$1(null);
};

/**
 * Keeps track of information needed to perform diffs for a given DOM node.
 * @param {!string} nodeName
 * @param {?string=} key
 * @constructor
 */
function NodeData(nodeName, key) {
  /**
   * The attributes and their values.
   * @const {!Object<string, *>}
   */
  this.attrs = createMap();

  /**
   * An array of attribute name/value pairs, used for quickly diffing the
   * incomming attributes to see if the DOM node's attributes need to be
   * updated.
   * @const {Array<*>}
   */
  this.attrsArr = [];

  /**
   * The incoming attributes for this Node, before they are updated.
   * @const {!Object<string, *>}
   */
  this.newAttrs = createMap();

  /**
   * The key used to identify this node, used to preserve DOM nodes when they
   * move within their parent.
   * @const
   */
  this.key = key;

  /**
   * Keeps track of children within this node by their key.
   * {?Object<string, !Element>}
   */
  this.keyMap = null;

  /**
   * Whether or not the keyMap is currently valid.
   * {boolean}
   */
  this.keyMapValid = true;

  /**
   * The node name for this node.
   * @const {string}
   */
  this.nodeName = nodeName;

  /**
   * @type {?string}
   */
  this.text = null;
}


/**
 * Initializes a NodeData object for a Node.
 *
 * @param {Node} node The node to initialize data for.
 * @param {string} nodeName The node name of node.
 * @param {?string=} key The key that identifies the node.
 * @return {!NodeData} The newly initialized data object
 */
var initData = function(node, nodeName, key) {
  var data = new NodeData(nodeName, key);
  node['__incrementalDOMData'] = data;
  return data;
};


/**
 * Retrieves the NodeData object for a Node, creating it if necessary.
 *
 * @param {Node} node The node to retrieve the data for.
 * @return {!NodeData} The NodeData for this Node.
 */
var getData = function(node) {
  var data = node['__incrementalDOMData'];

  if (!data) {
    var nodeName = node.nodeName.toLowerCase();
    var key = null;

    if (node instanceof Element) {
      key = node.getAttribute('key');
    }

    data = initData(node, nodeName, key);
  }

  return data;
};

/**
 * Copyright 2015 The Incremental DOM Authors. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS-IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

var symbols = {
  default: '__default',

  placeholder: '__placeholder'
};

/**
 * Applies an attribute or property to a given Element. If the value is null
 * or undefined, it is removed from the Element. Otherwise, the value is set
 * as an attribute.
 * @param {!Element} el
 * @param {string} name The attribute's name.
 * @param {?(boolean|number|string)=} value The attribute's value.
 */
var applyAttr = function(el, name, value) {
  if (value == null) {
    el.removeAttribute(name);
  } else {
    el.setAttribute(name, value);
  }
};

/**
 * Applies a property to a given Element.
 * @param {!Element} el
 * @param {string} name The property's name.
 * @param {*} value The property's value.
 */
var applyProp = function(el, name, value) {
  el[name] = value;
};


/**
 * Applies a style to an Element. No vendor prefix expansion is done for
 * property names/values.
 * @param {!Element} el
 * @param {string} name The attribute's name.
 * @param {*} style The style to set. Either a string of css or an object
 *     containing property-value pairs.
 */
var applyStyle = function(el, name, style) {
  if (typeof style === 'string') {
    el.style.cssText = style;
  } else {
    el.style.cssText = '';
    var elStyle = el.style;
    var obj = /** @type {!Object<string,string>} */(style);

    for (var prop in obj) {
      if (has(obj, prop)) {
        elStyle[prop] = obj[prop];
      }
    }
  }
};


/**
 * Updates a single attribute on an Element.
 * @param {!Element} el
 * @param {string} name The attribute's name.
 * @param {*} value The attribute's value. If the value is an object or
 *     function it is set on the Element, otherwise, it is set as an HTML
 *     attribute.
 */
var applyAttributeTyped = function(el, name, value) {
  var type = typeof value;

  if (type === 'object' || type === 'function') {
    applyProp(el, name, value);
  } else {
    applyAttr(el, name, /** @type {?(boolean|number|string)} */(value));
  }
};


/**
 * Calls the appropriate attribute mutator for this attribute.
 * @param {!Element} el
 * @param {string} name The attribute's name.
 * @param {*} value The attribute's value.
 */
var updateAttribute = function(el, name, value) {
  var data = getData(el);
  var attrs = data.attrs;

  if (attrs[name] === value) {
    return;
  }

  var mutator = attributes[name] || attributes[symbols.default];
  mutator(el, name, value);

  attrs[name] = value;
};


/**
 * A publicly mutable object to provide custom mutators for attributes.
 * @const {!Object<string, function(!Element, string, *)>}
 */
var attributes = createMap();

// Special generic mutator that's called for any attribute that does not
// have a specific mutator.
attributes[symbols.default] = applyAttributeTyped;

attributes[symbols.placeholder] = function() {};

attributes['style'] = applyStyle;

/**
 * Copyright 2015 The Incremental DOM Authors. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS-IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/** */
var notifications = {
  /**
   * Called after patch has compleated with any Nodes that have been created
   * and added to the DOM.
   * @type {?function(Array<!Node>)}
   */
  nodesCreated: null,

  /**
   * Called after patch has compleated with any Nodes that have been removed
   * from the DOM.
   * Note it's an applications responsibility to handle any childNodes.
   * @type {?function(Array<!Node>)}
   */
  nodesDeleted: null
};

/**
 * Gets the namespace to create an element (of a given tag) in.
 * @param {string} tag The tag to get the namespace for.
 * @param {?Node} parent
 * @return {?string} The namespace to create the tag in.
 */
var getNamespaceForTag = function(tag, parent) {
  if (tag === 'svg') {
    return 'http://www.w3.org/2000/svg';
  }

  if (getData(parent).nodeName === 'foreignObject') {
    return null;
  }

  return parent.namespaceURI;
};


/**
 * Creates an Element.
 * @param {Document} doc The document with which to create the Element.
 * @param {?Node} parent
 * @param {string} tag The tag for the Element.
 * @param {?string=} key A key to identify the Element.
 * @param {?Array<*>=} statics An array of attribute name/value pairs of the
 *     static attributes for the Element.
 * @return {!Element}
 */
var createElement = function(doc, parent, tag, key, statics) {
  var namespace = getNamespaceForTag(tag, parent);
  var el;

  if (namespace) {
    el = doc.createElementNS(namespace, tag);
  } else {
    el = doc.createElement(tag);
  }

  initData(el, tag, key);

  if (statics) {
    for (var i = 0; i < statics.length; i += 2) {
      updateAttribute(el, /** @type {!string}*/(statics[i]), statics[i + 1]);
    }
  }

  return el;
};


/**
 * Creates a Text Node.
 * @param {Document} doc The document with which to create the Element.
 * @return {!Text}
 */
var createText = function(doc) {
  var node = doc.createTextNode('');
  initData(node, '#text', null);
  return node;
};


/**
 * Creates a mapping that can be used to look up children using a key.
 * @param {?Node} el
 * @return {!Object<string, !Element>} A mapping of keys to the children of the
 *     Element.
 */
var createKeyMap = function(el) {
  var map = createMap();
  var children = el.children;
  var count = children.length;

  for (var i = 0; i < count; i += 1) {
    var child = children[i];
    var key = getData(child).key;

    if (key) {
      map[key] = child;
    }
  }

  return map;
};


/**
 * Retrieves the mapping of key to child node for a given Element, creating it
 * if necessary.
 * @param {?Node} el
 * @return {!Object<string, !Node>} A mapping of keys to child Elements
 */
var getKeyMap = function(el) {
  var data = getData(el);

  if (!data.keyMap) {
    data.keyMap = createKeyMap(el);
  }

  return data.keyMap;
};


/**
 * Retrieves a child from the parent with the given key.
 * @param {?Node} parent
 * @param {?string=} key
 * @return {?Node} The child corresponding to the key.
 */
var getChild = function(parent, key) {
  return key ? getKeyMap(parent)[key] : null;
};


/**
 * Registers an element as being a child. The parent will keep track of the
 * child using the key. The child can be retrieved using the same key using
 * getKeyMap. The provided key should be unique within the parent Element.
 * @param {?Node} parent The parent of child.
 * @param {string} key A key to identify the child with.
 * @param {!Node} child The child to register.
 */
var registerChild = function(parent, key, child) {
  getKeyMap(parent)[key] = child;
};

/**
 * Keeps track of the state of a patch.
 * @constructor
 */
function Context$1() {
  /**
   * @type {(Array<!Node>|undefined)}
   */
  this.created = notifications.nodesCreated && [];

  /**
   * @type {(Array<!Node>|undefined)}
   */
  this.deleted = notifications.nodesDeleted && [];
}


/**
 * @param {!Node} node
 */
Context$1.prototype.markCreated = function(node) {
  if (this.created) {
    this.created.push(node);
  }
};


/**
 * @param {!Node} node
 */
Context$1.prototype.markDeleted = function(node) {
  if (this.deleted) {
    this.deleted.push(node);
  }
};


/**
 * Notifies about nodes that were created during the patch opearation.
 */
Context$1.prototype.notifyChanges = function() {
  if (this.created && this.created.length > 0) {
    notifications.nodesCreated(this.created);
  }

  if (this.deleted && this.deleted.length > 0) {
    notifications.nodesDeleted(this.deleted);
  }
};

/**
 * Copyright 2015 The Incremental DOM Authors. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS-IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


/**
  * Keeps track whether or not we are in an attributes declaration (after
  * elementOpenStart, but before elementOpenEnd).
  * @type {boolean}
  */
var inAttributes = false;


/**
  * Keeps track whether or not we are in an element that should not have its
  * children cleared.
  * @type {boolean}
  */
var inSkip = false;


/**
 * Makes sure that there is a current patch context.
 * @param {*} context
 */
var assertInPatch = function(context) {
  if (!context) {
    throw new Error('Cannot call currentElement() unless in patch');
  }
};


/**
* Makes sure that keyed Element matches the tag name provided.
* @param {!string} nodeName The nodeName of the node that is being matched.
* @param {string=} tag The tag name of the Element.
* @param {?string=} key The key of the Element.
*/
var assertKeyedTagMatches = function(nodeName, tag, key) {
  if (nodeName !== tag) {
    throw new Error('Was expecting node with key "' + key + '" to be a ' +
        tag + ', not a ' + nodeName + '.');
  }
};


/**
 * Makes sure that a patch closes every node that it opened.
 * @param {?Node} openElement
 * @param {!Node|!DocumentFragment} root
 */
var assertNoUnclosedTags = function(openElement, root) {
  if (openElement === root) {
    return;
  }

  var currentElement = openElement;
  var openTags = [];
  while (currentElement && currentElement !== root) {
    openTags.push(currentElement.nodeName.toLowerCase());
    currentElement = currentElement.parentNode;
  }

  throw new Error('One or more tags were not closed:\n' +
      openTags.join('\n'));
};


/**
 * Makes sure that the caller is not where attributes are expected.
 * @param {string} functionName
 */
var assertNotInAttributes = function(functionName) {
  if (inAttributes) {
    throw new Error(functionName + '() may not be called between ' +
        'elementOpenStart() and elementOpenEnd().');
  }
};


/**
 * Makes sure that the caller is not inside an element that has declared skip.
 * @param {string} functionName
 */
var assertNotInSkip = function(functionName) {
  if (inSkip) {
    throw new Error(functionName + '() may not be called inside an element ' +
        'that has called skip().');
  }
};


/**
 * Makes sure that the caller is where attributes are expected.
 * @param {string} functionName
 */
var assertInAttributes = function(functionName) {
  if (!inAttributes) {
    throw new Error(functionName + '() must be called after ' +
        'elementOpenStart().');
  }
};


/**
 * Makes sure the patch closes virtual attributes call
 */
var assertVirtualAttributesClosed = function() {
  if (inAttributes) {
    throw new Error('elementOpenEnd() must be called after calling ' +
        'elementOpenStart().');
  }
};


/**
  * Makes sure that placeholders have a key specified. Otherwise, conditional
  * placeholders and conditional elements next to placeholders will cause
  * placeholder elements to be re-used as non-placeholders and vice versa.
  * @param {string} key
  */
var assertPlaceholderKeySpecified = function(key) {
  if (!key) {
    throw new Error('Placeholder elements must have a key specified.');
  }
};


/**
  * Makes sure that tags are correctly nested.
  * @param {string} nodeName
  * @param {string} tag
  */
var assertCloseMatchesOpenTag = function(nodeName, tag) {
  if (nodeName !== tag) {
    throw new Error('Received a call to close ' + tag + ' but ' +
        nodeName + ' was open.');
  }
};


/**
 * Makes sure that no children elements have been declared yet in the current
 * element.
 * @param {string} functionName
 * @param {?Node} previousNode
 */
var assertNoChildrenDeclaredYet = function(functionName, previousNode) {
  if (previousNode !== null) {
    throw new Error(functionName + '() must come before any child ' +
        'declarations inside the current element.');
  }
};


/**
 * Updates the state of being in an attribute declaration.
 * @param {boolean} value
 * @return {boolean} the previous value.
 */
var setInAttributes = function(value) {
  var previous = inAttributes;
  inAttributes = value;
  return previous;
};


/**
 * Updates the state of being in a skip element.
 * @param {boolean} value
 * @return {boolean} the previous value.
 */
var setInSkip = function(value) {
  var previous = inSkip;
  inSkip = value;
  return previous;
};

/** @type {?Context} */
var context = null;

/** @type {?Node} */
var currentNode;

/** @type {?Node} */
var currentParent;

/** @type {?Node} */
var previousNode;

/** @type {?Element|?DocumentFragment} */
var root;

/** @type {?Document} */
var doc;


/**
 * Patches the document starting at el with the provided function. This function
 * may be called during an existing patch operation.
 * @param {!Element|!DocumentFragment} node The Element or Document
 *     to patch.
 * @param {!function(T)} fn A function containing elementOpen/elementClose/etc.
 *     calls that describe the DOM.
 * @param {T=} data An argument passed to fn to represent DOM state.
 * @template T
 */
var patch = function(node, fn, data) {
  var prevContext = context;
  var prevRoot = root;
  var prevDoc = doc;
  var prevCurrentNode = currentNode;
  var prevCurrentParent = currentParent;
  var prevPreviousNode = previousNode;
  var previousInAttributes = false;
  var previousInSkip = false;

  context = new Context$1();
  root = node;
  doc = node.ownerDocument;
  currentNode = node;
  currentParent = null;
  previousNode = null;

  if (process.env.NODE_ENV !== 'production') {
    previousInAttributes = setInAttributes(false);
    previousInSkip = setInSkip(false);
  }

  enterNode();
  fn(data);
  exitNode();

  if (process.env.NODE_ENV !== 'production') {
    assertVirtualAttributesClosed();
    assertNoUnclosedTags(previousNode, node);
    setInAttributes(previousInAttributes);
    setInSkip(previousInSkip);
  }

  context.notifyChanges();

  context = prevContext;
  root = prevRoot;
  doc = prevDoc;
  currentNode = prevCurrentNode;
  currentParent = prevCurrentParent;
  previousNode = prevPreviousNode;
};


/**
 * Checks whether or not the current node matches the specified nodeName and
 * key.
 *
 * @param {?string} nodeName The nodeName for this node.
 * @param {?string=} key An optional key that identifies a node.
 * @return {boolean} True if the node matches, false otherwise.
 */
var matches$1 = function(nodeName, key) {
  var data = getData(currentNode);

  // Key check is done using double equals as we want to treat a null key the
  // same as undefined. This should be okay as the only values allowed are
  // strings, null and undefined so the == semantics are not too weird.
  return nodeName === data.nodeName && key == data.key;
};


/**
 * Aligns the virtual Element definition with the actual DOM, moving the
 * corresponding DOM node to the correct location or creating it if necessary.
 * @param {string} nodeName For an Element, this should be a valid tag string.
 *     For a Text, this should be #text.
 * @param {?string=} key The key used to identify this element.
 * @param {?Array<*>=} statics For an Element, this should be an array of
 *     name-value pairs.
 */
var alignWithDOM = function(nodeName, key, statics) {
  if (currentNode && matches$1(nodeName, key)) {
    return;
  }

  var node;

  // Check to see if the node has moved within the parent.
  if (key) {
    node = getChild(currentParent, key);
    if (node && process.env.NODE_ENV !== 'production') {
      assertKeyedTagMatches(getData(node).nodeName, nodeName, key);
    }
  }

  // Create the node if it doesn't exist.
  if (!node) {
    if (nodeName === '#text') {
      node = createText(doc);
    } else {
      node = createElement(doc, currentParent, nodeName, key, statics);
    }

    if (key) {
      registerChild(currentParent, key, node);
    }

    context.markCreated(node);
  }

  // If the node has a key, remove it from the DOM to prevent a large number
  // of re-orders in the case that it moved far or was completely removed.
  // Since we hold on to a reference through the keyMap, we can always add it
  // back.
  if (currentNode && getData(currentNode).key) {
    currentParent.replaceChild(node, currentNode);
    getData(currentParent).keyMapValid = false;
  } else {
    currentParent.insertBefore(node, currentNode);
  }

  currentNode = node;
};


/**
 * Clears out any unvisited Nodes, as the corresponding virtual element
 * functions were never called for them.
 */
var clearUnvisitedDOM = function() {
  var node = currentParent;
  var data = getData(node);
  var keyMap = data.keyMap;
  var keyMapValid = data.keyMapValid;
  var child = node.lastChild;
  var key;

  if (child === previousNode && keyMapValid) {
    return;
  }

  if (data.attrs[symbols.placeholder] && node !== root) {
    return;
  }

  while (child !== previousNode) {
    node.removeChild(child);
    context.markDeleted(/** @type {!Node}*/(child));

    key = getData(child).key;
    if (key) {
      delete keyMap[key];
    }
    child = node.lastChild;
  }

  // Clean the keyMap, removing any unusued keys.
  if (!keyMapValid) {
    for (key in keyMap) {
      child = keyMap[key];
      if (child.parentNode !== node) {
        context.markDeleted(child);
        delete keyMap[key];
      }
    }

    data.keyMapValid = true;
  }
};


/**
 * Changes to the first child of the current node.
 */
var enterNode = function() {
  currentParent = currentNode;
  currentNode = currentNode.firstChild;
  previousNode = null;
};


/**
 * Changes to the next sibling of the current node.
 */
var nextNode = function() {
  previousNode = currentNode;
  currentNode = currentNode.nextSibling;
};


/**
 * Changes to the parent of the current node, removing any unvisited children.
 */
var exitNode = function() {
  clearUnvisitedDOM();

  previousNode = currentParent;
  currentNode = currentParent.nextSibling;
  currentParent = currentParent.parentNode;
};


/**
 * Makes sure that the current node is an Element with a matching tagName and
 * key.
 *
 * @param {string} tag The element's tag.
 * @param {?string=} key The key used to identify this element. This can be an
 *     empty string, but performance may be better if a unique value is used
 *     when iterating over an array of items.
 * @param {?Array<*>=} statics An array of attribute name/value pairs of the
 *     static attributes for the Element. These will only be set once when the
 *     Element is created.
 * @return {!Element} The corresponding Element.
 */
var coreElementOpen = function(tag, key, statics) {
  alignWithDOM(tag, key, statics);
  enterNode();
  return /** @type {!Element} */(currentParent);
};


/**
 * Closes the currently open Element, removing any unvisited children if
 * necessary.
 *
 * @return {!Element} The corresponding Element.
 */
var coreElementClose = function() {
  if (process.env.NODE_ENV !== 'production') {
    setInSkip(false);
  }

  exitNode();
  return /** @type {!Element} */(previousNode);
};


/**
 * Makes sure the current node is a Text node and creates a Text node if it is
 * not.
 *
 * @return {!Text} The corresponding Text Node.
 */
var coreText = function() {
  alignWithDOM('#text', null, null);
  nextNode();
  return /** @type {!Text} */(previousNode);
};


/**
 * Gets the current Element being patched.
 * @return {!Element}
 */
var currentElement = function() {
  if (process.env.NODE_ENV !== 'production') {
    assertInPatch(context);
    assertNotInAttributes('currentElement');
  }
  return /** @type {!Element} */(currentParent);
};


/**
 * Skips the children in a subtree, allowing an Element to be closed without
 * clearing out the children.
 */
var skip = function() {
  if (process.env.NODE_ENV !== 'production') {
    assertNoChildrenDeclaredYet('skip', previousNode);
    setInSkip(true);
  }
  previousNode = currentParent.lastChild;
};

/**
 * The offset in the virtual element declaration where the attributes are
 * specified.
 * @const
 */
var ATTRIBUTES_OFFSET = 3;


/**
 * Builds an array of arguments for use with elementOpenStart, attr and
 * elementOpenEnd.
 * @const {Array<*>}
 */
var argsBuilder = [];


/**
 * @param {string} tag The element's tag.
 * @param {?string=} key The key used to identify this element. This can be an
 *     empty string, but performance may be better if a unique value is used
 *     when iterating over an array of items.
 * @param {?Array<*>=} statics An array of attribute name/value pairs of the
 *     static attributes for the Element. These will only be set once when the
 *     Element is created.
 * @param {...*} var_args Attribute name/value pairs of the dynamic attributes
 *     for the Element.
 * @return {!Element} The corresponding Element.
 */
var elementOpen = function(tag, key, statics, var_args) {
  if (process.env.NODE_ENV !== 'production') {
    assertNotInAttributes('elementOpen');
    assertNotInSkip('elementOpen');
  }

  var node = coreElementOpen(tag, key, statics);
  var data = getData(node);

  /*
   * Checks to see if one or more attributes have changed for a given Element.
   * When no attributes have changed, this is much faster than checking each
   * individual argument. When attributes have changed, the overhead of this is
   * minimal.
   */
  var attrsArr = data.attrsArr;
  var newAttrs = data.newAttrs;
  var attrsChanged = false;
  var i = ATTRIBUTES_OFFSET;
  var j = 0;

  for (; i < arguments.length; i += 1, j += 1) {
    if (attrsArr[j] !== arguments[i]) {
      attrsChanged = true;
      break;
    }
  }

  for (; i < arguments.length; i += 1, j += 1) {
    attrsArr[j] = arguments[i];
  }

  if (j < attrsArr.length) {
    attrsChanged = true;
    attrsArr.length = j;
  }

  /*
   * Actually perform the attribute update.
   */
  if (attrsChanged) {
    for (i = ATTRIBUTES_OFFSET; i < arguments.length; i += 2) {
      newAttrs[arguments[i]] = arguments[i + 1];
    }

    for (var attr in newAttrs) {
      updateAttribute(node, attr, newAttrs[attr]);
      newAttrs[attr] = undefined;
    }
  }

  return node;
};


/**
 * Declares a virtual Element at the current location in the document. This
 * corresponds to an opening tag and a elementClose tag is required. This is
 * like elementOpen, but the attributes are defined using the attr function
 * rather than being passed as arguments. Must be folllowed by 0 or more calls
 * to attr, then a call to elementOpenEnd.
 * @param {string} tag The element's tag.
 * @param {?string=} key The key used to identify this element. This can be an
 *     empty string, but performance may be better if a unique value is used
 *     when iterating over an array of items.
 * @param {?Array<*>=} statics An array of attribute name/value pairs of the
 *     static attributes for the Element. These will only be set once when the
 *     Element is created.
 */
var elementOpenStart = function(tag, key, statics) {
  if (process.env.NODE_ENV !== 'production') {
    assertNotInAttributes('elementOpenStart');
    setInAttributes(true);
  }

  argsBuilder[0] = tag;
  argsBuilder[1] = key;
  argsBuilder[2] = statics;
};


/***
 * Defines a virtual attribute at this point of the DOM. This is only valid
 * when called between elementOpenStart and elementOpenEnd.
 *
 * @param {string} name
 * @param {*} value
 */
var attr = function(name, value) {
  if (process.env.NODE_ENV !== 'production') {
    assertInAttributes('attr');
  }

  argsBuilder.push(name, value);
};


/**
 * Closes an open tag started with elementOpenStart.
 * @return {!Element} The corresponding Element.
 */
var elementOpenEnd = function() {
  if (process.env.NODE_ENV !== 'production') {
    assertInAttributes('elementOpenEnd');
    setInAttributes(false);
  }

  var node = elementOpen.apply(null, argsBuilder);
  argsBuilder.length = 0;
  return node;
};


/**
 * Closes an open virtual Element.
 *
 * @param {string} tag The element's tag.
 * @return {!Element} The corresponding Element.
 */
var elementClose = function(tag) {
  if (process.env.NODE_ENV !== 'production') {
    assertNotInAttributes('elementClose');
  }

  var node = coreElementClose();

  if (process.env.NODE_ENV !== 'production') {
    assertCloseMatchesOpenTag(getData(node).nodeName, tag);
  }

  return node;
};


/**
 * Declares a virtual Element at the current location in the document that has
 * no children.
 * @param {string} tag The element's tag.
 * @param {?string=} key The key used to identify this element. This can be an
 *     empty string, but performance may be better if a unique value is used
 *     when iterating over an array of items.
 * @param {?Array<*>=} statics An array of attribute name/value pairs of the
 *     static attributes for the Element. These will only be set once when the
 *     Element is created.
 * @param {...*} var_args Attribute name/value pairs of the dynamic attributes
 *     for the Element.
 * @return {!Element} The corresponding Element.
 */
var elementVoid = function(tag, key, statics, var_args) {
  var node = elementOpen.apply(null, arguments);
  elementClose.apply(null, arguments);
  return node;
};


/**
 * Declares a virtual Element at the current location in the document that is a
 * placeholder element. Children of this Element can be manually managed and
 * will not be cleared by the library.
 *
 * A key must be specified to make sure that this node is correctly preserved
 * across all conditionals.
 *
 * @param {string} tag The element's tag.
 * @param {string} key The key used to identify this element.
 * @param {?Array<*>=} statics An array of attribute name/value pairs of the
 *     static attributes for the Element. These will only be set once when the
 *     Element is created.
 * @param {...*} var_args Attribute name/value pairs of the dynamic attributes
 *     for the Element.
 * @return {!Element} The corresponding Element.
 */
var elementPlaceholder = function(tag, key, statics, var_args) {
  if (process.env.NODE_ENV !== 'production') {
    assertPlaceholderKeySpecified(key);
  }

  elementOpen.apply(null, arguments);
  skip();
  return elementClose.apply(null, arguments);
};


/**
 * Declares a virtual Text at this point in the document.
 *
 * @param {string|number|boolean} value The value of the Text.
 * @param {...(function((string|number|boolean)):string)} var_args
 *     Functions to format the value which are called only when the value has
 *     changed.
 * @return {!Text} The corresponding text node.
 */
var text = function(value, var_args) {
  if (process.env.NODE_ENV !== 'production') {
    assertNotInAttributes('text');
    assertNotInSkip('text');
  }

  var node = coreText();
  var data = getData(node);

  if (data.text !== value) {
    data.text = /** @type {string} */(value);

    var formatted = value;
    for (var i = 1; i < arguments.length; i += 1) {
      formatted = arguments[i](formatted);
    }

    node.data = formatted;
  }

  return node;
};

function forEach(val, fn, ctx) {
	if (val && typeof val.forEach === "function") val.forEach(fn, ctx);else _.forEach(val, fn, ctx);
}

function map(val, fn, ctx) {
	var res = [];
	forEach(val, function (v, i, l) {
		res.push(fn.call(ctx, v, i, l));
	});
	return res;
}

// like jQuery's empty(), removes all children
function emptyNode(node) {
	while (node.lastChild) {
		node.removeChild(node.lastChild);
	}return node;
}

var entity = /&(?:#x[a-f0-9]+|#[0-9]+|[a-z0-9]+);?/ig;

function containsEntities(str) {
	return entity.test(str);
}

// converts html entities to unicode
var decodeEntities = function () {
	if (typeof document === "undefined") return;

	// this prevents any overhead from creating the object each time
	var element = document.createElement('div');

	return function decodeHTMLEntities(str) {
		str = str.replace(entity, function (m) {
			element.innerHTML = m;
			return element.textContent;
		});

		emptyNode(element);

		return str;
	};
}();

// convert html into DOM nodes
function toString(value) {
	if (typeof value !== "string") {
		value = value != null ? value.toString() : "";
	}

	return value;
}

function joinValues() {
	return _.toArray(arguments).map(toString).join("");
}

function matches(node, selector) {
	if (_.isArray(selector)) return selector.some(function (s) {
		return matches(node, s);
	});

	if (selector instanceof window.Node) {
		return node === selector;
	}

	if (typeof selector === "function") {
		return !!selector(node);
	}

	if (node.nodeType === window.Node.ELEMENT_NODE) {
		return matchesSelector(node, selector);
	}

	return false;
}

function getPropertyFromClass(obj, prop) {
	var val;
	var proto = Object.getPrototypeOf(obj);

	if (typeof prop === "string" && prop) {
		(function () {
			var p = prop;
			prop = function prop(c) {
				return c[p];
			};
		})();
	}

	if (typeof prop !== "function") {
		throw new Error("Expecting function or string for property.");
	}

	while (proto) {
		val = merge.defaults(val, prop(proto.constructor));
		proto = Object.getPrototypeOf(proto);
	}

	return val;
}

var utils = Object.freeze({
	assign: _.assign,
	forEach: forEach,
	map: map,
	emptyNode: emptyNode,
	containsEntities: containsEntities,
	decodeEntities: decodeEntities,
	toString: toString,
	joinValues: joinValues,
	matches: matches,
	getPropertyFromClass: getPropertyFromClass
});

function autotext(fn, that) {
	var node = coreText();

	function renderText(c) {
		c.node = node;
		var data = getData(node);
		var value = fn.call(that, node);

		if (data.text !== value) {
			data.text = value;
			node.data = toString(value);
		}
	}

	return Trackr.autorun(renderText);
}

function autoelement(node, fn) {
	return Trackr.autorun(function (c) {
		c.element = node;

		// catch errors that idom likes to throw
		var cur = undefined;
		try {
			cur = currentElement();
		} catch (e) {}

		// call straight or patch
		if (cur === node) fn(node, c);else patch(node, function () {
			return fn(node, c);
		});
	});
}

var execDeletedCallbacks = function execDeletedCallbacks(node) {
	var data = getData(node);
	var cbs = data.deletedCallbacks;
	if (!cbs) return;
	cbs.splice(0, cbs.length).forEach(function (fn) {
		fn.call(null, node);
	});
};

notifications.nodesDeleted = function (nodes) {
	_.forEach(nodes, function (node) {
		if (node.childNodes) notifications.nodesDeleted(node.childNodes);
		execDeletedCallbacks(node);
	});
};

function onDestroy(node, fn) {
	if (typeof fn !== "function") {
		throw new Error("Expecting function.");
	}

	var data = getData(node);
	if (!data.deletedCallbacks) data.deletedCallbacks = [];
	data.deletedCallbacks.push(fn);
}

function removeDestroyListener(node, fn) {
	var data = getData(node);
	if (data.deletedCallbacks) {
		data.deletedCallbacks = _.without(data.deletedCallbacks, fn);
	}
}

// function renderHTML(node) {
// 	if (node.type == "text") {
// 		return idom.text(node.content);
// 	}
//
// 	if (node.type == "tag") {
// 		var argsArray = [node.name, null, null];
//
// 		// convert attrs into a flat array
// 		for (var attr in node.attrs) {
// 			argsArray.push(attr);
// 			argsArray.push(node.attrs[attr]);
// 		}
//
// 		if (node.voidElement) {
// 			return idom.elementVoid.apply(idom, argsArray);
// 		} else {
// 			idom.elementOpen.apply(idom, argsArray);
//
// 			for (var i = 0; i < node.children.length; i++) {
// 				renderHTML(node.children[i]);
// 			}
//
// 			return idom.elementClose(node.name);
// 		}
// 	}
// }

// export function html(src) {
// 	var tree = parseHTML("<div>" + src + "</div>");
// 	if (_.isArray(tree)) tree = tree[0];
// 	return tree.children.map(renderHTML);
// }

function updateProperty(el, name, value) {
	var data = getData(el);
	var attrs = data.attrs;

	if (attrs[name] === value) {
		return;
	}

	el[name] = value;
	attrs[name] = value;
}

var idom = Object.freeze({
	updateAttribute: updateAttribute,
	getData: getData,
	autotext: autotext,
	autoelement: autoelement,
	onDestroy: onDestroy,
	removeDestroyListener: removeDestroyListener,
	updateProperty: updateProperty,
	patch: patch,
	currentElement: currentElement,
	skip: skip,
	elementVoid: elementVoid,
	elementOpenStart: elementOpenStart,
	elementOpenEnd: elementOpenEnd,
	elementOpen: elementOpen,
	elementClose: elementClose,
	elementPlaceholder: elementPlaceholder,
	text: text,
	attr: attr,
	symbols: symbols,
	attributes: attributes,
	applyAttr: applyAttr,
	applyProp: applyProp,
	notifications: notifications
});

var templates = {};
var types = {};

function Template(name, render) {
	if (typeof name !== "string" || name === "") {
		throw new Error("Expecting a non-empty string for template name.");
	}

	if (typeof render !== "function") {
		throw new Error("Expecting a function for render.");
	}

	events.EventEmitter.call(this);
	this.setMaxListeners(0);

	// internal state
	this.s = {
		// template name
		name: name,
		// render method
		render: render,
		// dependency for easy invalidate
		renderdep: new Trackr.Dependency(),
		// holds the helpers
		helpers: new trackrObjects.Map()
	};

	// default plugins
	this.use("decorators");
	this.initialize();
}

Template.extend = subclass;
Template.prototype = Object.create(events.EventEmitter.prototype);
Template.prototype.constructor = Template;
Template.prototype.type = "template";
Template.prototype.initialize = function () {};

// plugin proxy for contexts
Template.prototype.use = function use(p) {
	return load(this, p, _.toArray(arguments).slice(1));
};

Template.prototype.createContext = function (data, parent) {
	if (data instanceof Context) {
		;
		var _ref = [data, null];
		parent = _ref[0];
		data = _ref[1];
	}var ctx = new Context(data, parent, this);
	this.emit("context", ctx);
	return ctx;
};

Template.prototype.invalidate = function () {
	this.s.renderdep.changed();
	return this;
};

Template.prototype.render = function (data, key) {
	this.s.renderdep.depend();
	var ctx = this.createContext(data);
	this.s.render(ctx, key);
	this.emit("render", ctx);
	return ctx;
};

Template.prototype.paint = function (node, data) {
	var _this = this;

	if (typeof node === "string") node = document.querySelector(node);
	if (!node || node.nodeType !== Node.ELEMENT_NODE) {
		throw new Error("Expecting a valid DOM element to paint.");
	}

	var ctx = data instanceof Context ? data : new Context(data);
	var c = Trackr.autorun(function () {
		patch(node, function () {
			return _this.render(ctx);
		});
	});

	c.context = ctx;
	c.template = this;

	c.onStop(function () {
		patch(node, function () {});
	});

	return c;
};

var globalHelpers = new trackrObjects.Map();

var helpers = Template.prototype.helpers = function (key, value) {
	var _this2 = this;

	if ((typeof key === "undefined" ? "undefined" : babelHelpers.typeof(key)) === "object") {
		Object.keys(key).forEach(function (k) {
			return helpers.call(_this2, k, key[k]);
		});
		return this;
	}

	var h = globalHelpers;
	if (this instanceof Template) h = this.s.helpers;

	if (typeof value === "undefined") h.delete(key);else h.set(key, value);

	return this;
};

var getHelper = Template.prototype.getHelper = function (key) {
	if (this instanceof Template) {
		return this.s.helpers.get(key);
	}

	return globalHelpers.get(key);
};

function render(name, ctx, key) {
	var tpl = getByName(name);
	if (tpl) return tpl.render(ctx, key);
}

function getByName(name) {
	return templates[name];
}

function paint(tpl, node, data) {
	if (typeof tpl === "string") {
		tpl = getByName(tpl);
		if (!tpl) throw new Error("No template exists with name '" + tpl + "'");
	}
	if (!(tpl instanceof Template)) {
		throw new Error("Expecting template name or instance of Template.");
	}
	return tpl.paint(node, data);
}

function registerType(type, props) {
	if (typeof type !== "string" || !type) {
		throw new Error("Expecting non-empty string for type.");
	}
	if (_.has(types, type)) {
		throw new Error("Template type '" + type + "' already exists.");
	}

	if (typeof props === "function") {
		props = { initialize: props };
	}

	props = _.assign({}, props, { type: type });

	var T = Template;
	if (props.extends && _.has(types, props.extends)) {
		T = types[props.extends];
		delete props.extends;
	}

	function wrap(k, f) {
		return function () {
			var osuper = this.super;
			this.super = T.prototype[k];
			var ret = f.apply(this, arguments);
			this.super = osuper;
			return ret;
		};
	}

	for (var k in props) {
		var fn = props[k];
		if (typeof fn !== "function") continue;
		if (/this\.super|this\["super"\]|this\['super'\]/.test(fn.toString())) {
			props[k] = wrap(k, fn);
		}
	}

	return types[type] = T.extend(props);
}

function create(name, type, render) {
	if (typeof name !== "string" || !name) {
		throw new Error("Expecting non-empty string for name.");
	}
	if (getByName(name)) {
		throw new Error("Template '" + name + "' already exists.");
	}

	if (typeof type === "function") {
		;
		var _ref2 = [type, null];
		render = _ref2[0];
		type = _ref2[1];
	}if (type && !_.has(types, type)) {
		throw new Error("Template type " + type + " does not exist.");
	}

	var T = type ? types[type] : Template;
	return templates[name] = new T(name, render);
}

function Context(data, parent, template) {
	if (!(this instanceof Context)) {
		return new Context(parent);
	}

	events.EventEmitter.call(this);
	this.setMaxListeners(0);

	if (data instanceof Context) {
		var _ref = [parent, data, null];
		template = _ref[0];
		parent = _ref[1];
		data = _ref[2];
	}

	if (parent && !(parent instanceof Context)) {
		throw new Error("Expecting an instance of Context for parent.");
	}

	// internal view state
	this.s = {
		// parent view for scope
		parent: parent || null,
		// associated template object
		template: template || null,
		// holds lexical data
		scope: {},
		// holds "this" data
		data: new trackrObjects.Variable()
	};

	// set initial data
	if (data != null) this.set(data);
	if (!this.parent) this.dataVar.set(data);
}

Context.prototype = Object.create(events.EventEmitter.prototype);
Context.prototype.constructor = Context;

assignProps(Context.prototype, {
	scope: function scope() {
		return this.s.scope;
	},
	dataVar: function dataVar() {
		return this.s.data;
	},
	data: function data() {
		return this.s.data.get();
	},
	template: function template() {
		return this.s.template;
	},
	parent: function parent() {
		return this.s.parent;
	}
});

Context.prototype.set = function (key, value) {
	var _this = this;

	if ((typeof key === "undefined" ? "undefined" : babelHelpers.typeof(key)) === "object") {
		_.forEach(key, function (v, k) {
			return _this.set(k, v);
		});
		return this;
	}

	if (key === "this") {
		this.s.data.set(value);
	} else {
		if (typeof value === "undefined") delete this.s.scope[key];else this.s.scope[key] = value;
	}

	return this;
};

Context.prototype.get = function (key) {
	return _.has(this.s.scope, key) ? this.s.scope[key] : void 0;
};

Context.prototype.parentData = function (dist) {
	if (typeof dist !== "number" || isNaN(dist)) dist = 1;

	if (dist >= 0) {
		var _v = this;
		while (dist && _v) {
			dist--;
			_v = _v.parent;
		}
		return !dist && _v ? _v.data : void 0;
	}

	var views = [];
	var v = this;

	while (v) {
		views.push(v);
		v = v.parent;
	}

	var view = views[Math.abs(dist) - 1];
	return view ? view.data : void 0;
};

Context.prototype.getTemplateContext = function () {
	var view = this;

	while (view) {
		if (view.template != null) {
			return view;
		}

		view = view.parent;
	}
};

Context.prototype.getTemplate = function () {
	var ctx = this.getTemplateContext();
	return ctx ? ctx.template : void 0;
};

var lookup = function lookup(tpl, ctx, key) {
	var view = undefined;

	// 0-0. check for this
	if (!key || key === "this") {
		view = ctx;
		while (view) {
			var _val = view.data;
			if (_val !== void 0) return _val;
			view = view.parent;
		}
	}

	// 0-1. special method $this
	if (key === "$this") {
		return function (d) {
			return ctx.parentData(d);
		};
	}

	// 1. check closest template helpers
	if (tpl) {
		var _val2 = tpl.getHelper(key);
		if (_val2 !== void 0) return _val2;
	}

	// 2. check lexical scope
	view = ctx;
	while (view) {
		var _val3 = view.get(key);
		if (_val3 !== void 0) return _val3;
		view = view.parent;
	}

	// 3. check global helpers
	var val = getHelper(key);
	if (val !== void 0) return val;
};

Context.prototype.lookup = function (key) {
	return lookup(this.getTemplate(), this, key);
};

// declare, but don't set the plugin variable
// exports and imports are hoisted, so this cannot do anything to the value
var plugins;

function load(tpl, plugin, args) {
	var name = undefined;
	if (plugins == null) plugins = {};

	if (_.isString(plugin)) {
		name = plugin;
		if (plugins[name] == null) throw new Error("No plugin exists with id '" + name + "'.");

		plugin = plugins[name];
	}

	if (!_.isFunction(plugin)) throw new Error("Expecting string or function for plugin");

	// check if plugin is already loaded on this template
	if (tpl._loaded_plugins == null) tpl._loaded_plugins = [];
	if (tpl._loaded_plugins.some(function (p) {
		return p.plugin === plugin;
	})) return tpl;

	if (args == null) args = [];
	if (!_.isArray(args)) args = [args];
	tpl._loaded_plugins.push({
		name: name,
		plugin: plugin,
		args: args
	});

	plugin.apply(tpl, args);
	return tpl;
}

function register(name, fn) {
	if (plugins == null) plugins = {};

	if (typeof name !== "string") {
		throw new Error("Expecting string name for plugin.");
	}

	if (typeof fn !== "function") {
		throw new Error("Expecting function for plugin.");
	}

	if (fn === plugins[name]) return;
	if (plugins[name] != null) {
		throw new Error("Refusing to overwrite existing plugin \"name\".");
	}

	plugins[name] = fn;
}

function get(name) {
	return plugins && plugins[name];
}



var plugins$1 = Object.freeze({
	load: load,
	register: register,
	get: get
});

var decorators = {};

function plugin() {
	this._decorators = {};
	this.decorate = add;
	this.stopDecorating = remove;
	this.findDecorator = find;
}

register("decorators", plugin);

// creates a decorator
function add(name, fn, options) {
	if ((typeof name === "undefined" ? "undefined" : babelHelpers.typeof(name)) === "object") {
		options = fn;
		_.each(name, function (fn, n) {
			if (_.isArray(fn)) add.call(this, n, fn[0], fn[1]);else add.call(this, n, fn, options);
		}, this);
		return this;
	}

	if (typeof name !== "string" || name === "") throw new Error("Expecting non-empty string for decorator name.");
	if (typeof fn !== "function") throw new Error("Expecting function for decorator.");

	var decs = this._decorators ? this._decorators : decorators;
	decs[name] = {
		template: this,
		callback: fn,
		options: options || {}
	};

	return this;
}

// removes a decorator
function remove(name) {
	if (this._decorators) {
		if (name == null) this._decorators = {};else delete this._decorators[name];
	} else if (name) {
		delete decorators[name];
	}

	return this;
}

function find(name) {
	var decs = this._decorators;
	if (decs != null && decs[name]) {
		return decs[name];
	}
}

// finds the best decorator
function lookup$1(ctx, name) {
	// look on the closest template
	var owner = ctx.getTemplateContext();
	if (owner) {
		var dec = owner.template.findDecorator(name);
		if (dec) return _.assign({ owner: owner }, dec);
	}

	// look globally
	if (decorators[name]) {
		return _.assign({ owner: global }, decorators[name]);
	}
}

function setAttribute(node, name, values) {
	var v = !values || !values.length || values[0] == null ? "" : values[0];
	updateAttribute(node, name, v);
}

function render$1(_ctx, name, value) {
	var node = currentElement();
	if (!node) throw new Error("Not currently patching.");

	// each decorator is given its own context
	var ctx = new Context(null, _ctx);

	var isStaticValue = typeof value !== "function";
	var getValue = function getValue() {
		var val = isStaticValue ? value : value(ctx);
		if (!_.isArray(val)) val = [val];
		return val;
	};

	// look up decorator by name
	var d = lookup$1(ctx, name);

	// quick escape if static value and no decorator
	if (isStaticValue && !d) {
		setAttribute(node, name, getValue());
		return;
	}

	var anim = undefined,
	    comp = undefined;
	var cancel = false;
	var decctx = _.assign({
		target: node,
		context: ctx
	}, d);

	var run = function run(c) {
		decctx.comp = c;
		var val = getValue();

		if (!d) {
			setAttribute(node, name, val);
			return;
		}

		// execute the callback
		d.callback.apply(d.template, [decctx].concat(val));
	};

	// defer computation if desired
	if (d && d.options && d.options.defer) {
		// clean up when current computation reruns
		var pcomp = Trackr.currentComputation;
		if (!pcomp) throw new Error("Can only render decorator in a computation.");
		pcomp.onInvalidate(function () {
			cancel = true;
			if (anim) raf.cancel(anim);
			if (comp) comp.stop();
		});

		// run decorator on the next frame
		anim = raf(function () {
			if (cancel) return;
			anim = null;
			comp = Trackr.autorun(run);
		});
	} else {
		Trackr.autorun(run);
	}
}

var decorators$1 = Object.freeze({
	plugin: plugin,
	default: plugin,
	add: add,
	remove: remove,
	find: find,
	lookup: lookup$1,
	render: render$1
});

var value_types = ["radio", "option"];
var selection_types = ["text", "search", "tel", "url", "password"];

var formBindings = {};

function plugin$2(options) {
	this.use("decorators");

	options = options || {};

	// add methods
	this._formBindings = {};
	this.twoway = this.addFormBinding = add$2;
	this.getFormBinding = get$1;
	this.removeFormBinding = remove$2;

	// add main binding decorator
	this.decorate("bind-to", bindTo.bind(this, options, false));
	this.decorate("lazybind-to", bindTo.bind(this, options, true));

	// add value decorator for radios and options
	this.decorate("value", valueOf, { parse: "string" });
}

register("twoway", plugin$2);

function add$2(id, getter, onChange) {
	if (_.isObject(id)) {
		_.each(id, function (v, k) {
			add$2.call(this, k, v);
		}, this);
		return this;
	}

	if (typeof id !== "string") throw new Error("Expecting a string for the form binding ID.");

	var bindings = this._formBindings || formBindings;
	if (bindings[id] != null) throw new Error("A form binding with id '" + id + "' already exists.");

	if (_.isObject(getter) && onChange == null) {
		onChange = getter.change;
		getter = getter.get;
	}

	if (typeof getter !== "function") throw new Error("Expecting a function or object for the form binding getter.");
	if (typeof onChange !== "function") onChange = null;

	bindings[id] = {
		get: getter,
		change: onChange
	};

	return this;
}

function get$1(id) {
	if (typeof id !== "string") return;

	if (this._formBindings && this._formBindings[id]) {
		return this._formBindings[id];
	}

	return formBindings[id];
}

function remove$2(id) {
	if (this._formBindings) {
		if (id == null) this._formBindings = {};else delete this._formBindings[id];
	} else if (id) {
		delete formBindings[id];
	}

	return this;
}

var type_map = {
	"text": ["text", "color", "email", "password", "search", "tel", "url", "hidden"],
	"number": ["number", "range"],
	"date": ["date", "datetime", "datetime-local", "month", "time", "week"],
	"file": ["file"],
	"checkbox": ["checkbox"],
	"radio": ["radio"]
};

function getType(el) {
	switch (el.tagName.toLowerCase()) {
		case "input":
			for (var type in type_map) {
				if (_.contains(type_map[type], el.type)) return type;
			}
			break;

		case "select":
			return "select";

		case "option":
			return "option";

		case "textarea":
			return "text";
	}
}

function getNodeValue(node, type) {
	if (type == null) type = getType(node);
	var val;

	switch (type) {
		case "number":
			val = node.valueAsNumber;
			break;
		case "text":
			val = node.value;
			break;

		case "checkbox":
			val = node.checked;
			break;

		case "date":
			val = node.valueAsDate;
			break;

		case "select":
			var opt = node.querySelector("option:checked");
			if (opt != null) val = opt.$bound_value;
			break;

		case "file":
			val = !node.multiple ? node.files[0] : _.toArray(node.files);
			break;

		case "radio":
			val = node.$bound_value;
			break;
	}

	return val;
}

function setLiveValue(live, el, fn) {
	var active = document.activeElement === el;
	if (!live && active) return;
	var sel = selection_types.indexOf(el.type) > -1;
	var pos = sel ? [el.selectionStart, el.selectionEnd] : null;
	fn();
	if (sel && active) el.setSelectionRange.apply(el, pos);
}

function setNodeValue(el, val, type, live) {
	if (type == null) type = getType(el);

	switch (type) {
		case "number":
			setLiveValue(live, el, function () {
				updateProperty(el, _.isNumber(val) ? "valueAsNumber" : "value", val);
			});
			break;

		case "text":
			setLiveValue(live, el, function () {
				updateProperty(el, "value", toString(val));
			});
			break;

		case "checkbox":
			updateProperty(el, "checked", Boolean(val));
			break;

		case "date":
			setLiveValue(live, el, function () {
				updateProperty(el, _.isDate(val) ? "valueAsDate" : "value", val);
			});
			break;

		case "select":
			_.toArray(el.querySelectorAll("option")).forEach(function (opt) {
				updateProperty(opt, "selected", opt.$bound_value === val);
			});
			break;

		case "radio":
			updateProperty(el, "checked", el.$bound_value === val);
			break;
	}
}

function bindTo(options, lazy, d, id) {
	var fbind = this.getFormBinding(id);
	if (fbind == null) return;

	var el = d.target;
	var args = _.toArray(arguments).slice(4);
	var type = getType(el);
	var onChange = undefined;
	var twctx = _.pick(d, "context", "template", "target");

	// detect changes to the input's value
	if (typeof fbind.change === "function") {
		onChange = function onChange(e) {
			fbind.change.apply(_.assign(twctx, {
				original: e
			}), [getNodeValue(el, type)].concat(args));
		};

		el.addEventListener("change", onChange);
		el.addEventListener("paste", onChange);
		if (!(options.lazy || lazy)) el.addEventListener("keyup", onChange);

		d.comp.onInvalidate(function () {
			el.removeEventListener("change", onChange);
			el.removeEventListener("paste", onChange);
			el.removeEventListener("keyup", onChange);
		});
	}

	var nodeValueComp = undefined,
	    stopped = false;

	d.comp.onInvalidate(function () {
		stopped = true;
		if (nodeValueComp) nodeValueComp.stop();
	});

	// reactively set the value on the input
	// deferred so value decorators run
	_.defer(function () {
		if (stopped) return;
		nodeValueComp = Trackr.autorun(function () {
			setNodeValue(el, fbind.get.apply(twctx, args), type, options.live);
		});
	});
}

function valueOf(d, val) {
	var el = d.target,
	    type = getType(el);

	if (!_.contains(value_types, type)) {
		updateAttribute(el, "value", val);
		return;
	}

	el.$bound_value = val;
	updateProperty(el, "value", val);
}

function plugin$3() {
	this.use("decorators");
	this.decorate("ref", decorator);
}

register("refs", plugin$3);

var warned = [];

function warnOnce(msg) {
	if (_.contains(warned, msg)) return;
	console.warn(msg);
	warned.push(msg);
}

function decorator(d, key) {
	var ctx = d.owner;
	if (!ctx.refs) ctx.refs = {};

	// warn about overwrites overwrite
	if (_.has(ctx.refs, key)) {
		warnOnce("Multiple elements with reference '" + key + "'.");
		return;
	}

	// set the reference
	ctx.refs[key] = d.target;

	// remove the reference when the element disappears
	d.comp.onInvalidate(function () {
		delete ctx.refs[key];
	});
}

// export function find(key) {
// 	var tpls = [ this ],
// 		tpl;
//
// 	while (tpls.length) {
// 		tpl = tpls.shift();
// 		if (tpl.refs && tpl.refs[key]) return tpl.refs[key];
// 		// tpls = tpls.concat(tpl.getComponents());
// 	}
//
// 	return null;
// }

var decorators$2 = {};
var actions = {};

// Action Class
var Action = function () {
	function Action(name, context) {
		babelHelpers.classCallCheck(this, Action);

		this.name = name;
		this.context = context;
		this.bubbles = true;
	}

	babelHelpers.createClass(Action, [{
		key: "stopPropagation",
		value: function stopPropagation() {
			this.bubbles = false;
			return this;
		}
	}], [{
		key: "create",
		value: function create(name, context) {
			var action = undefined;

			if (typeof name === "string") {
				action = new Action(name, context);
			} else if (_.isObject(name) && !(name instanceof Action)) {
				action = _.extend(new Action(), action);
			} else {
				action = name;
			}

			if (!(action instanceof Action)) {
				throw new Error("Expecting action name, object or instance of Action.");
			}

			return action;
		}
	}]);
	return Action;
}();

// the plugin
function plugin$1() {
	this.use("decorators");
	this._actions = {};
	this.actions = this.addAction = add$1;
	this.removeAction = remove$1;
	this.decorate(decorators$2);
}

register("actions", plugin$1);

// standard dom events
defineEvent(['load', 'scroll', 'click', 'dblclick', 'mousedown', 'mouseup', 'mouseenter', 'mouseleave', 'keydown', 'keypress', 'keyup', 'blur', 'focus', 'change', 'input', 'submit', 'reset', 'drag', 'dragdrop', 'dragend', 'dragenter', 'dragexit', 'dragleave', 'dragover', 'dragstart', 'drop']);

function defineEvent(event) {
	if (_.isArray(event)) return event.forEach(defineEvent);

	decorators$2["on-" + event] = function (decor, key) {
		var node = undefined,
		    args = undefined;
		var listener = function listener(e) {
			// create a new action object
			var action = new Action(key, decor.context);
			action.original = e;
			action.target = action.node = node;
			action.owner = decor.owner;
			action.template = decor.template;

			// fire the action
			fire(action, null, args);
		};

		node = decor.target;
		args = Array.prototype.slice.call(arguments, 2);

		node.addEventListener(event, listener);
		decor.comp.onInvalidate(function () {
			node.removeEventListener(event, listener);
		});
	};
}

// Msutache Instance Methods
function add$1(name, fn) {
	if ((typeof name === "undefined" ? "undefined" : babelHelpers.typeof(name)) === "object" && fn == null) {
		_.each(name, function (fn, n) {
			add$1.call(this, n, fn);
		}, this);
		return this;
	}

	if (typeof name !== "string" || name === "") throw new Error("Expecting non-empty string for action name.");
	if (typeof fn !== "function") throw new Error("Expecting function for action.");

	var obj = this._actions;
	if (!obj) obj = actions;
	if (obj[name] == null) obj[name] = [];
	if (! ~obj[name].indexOf(fn)) obj[name].push(fn);

	return this;
}

function remove$1(name, fn) {
	if (typeof name === "function" && fn == null) {
		fn = name;
		name = null;
	}

	var obj = this._actions;
	if (!obj) obj = actions;

	if (name == null && fn == null) {
		// clear actions, but never on the global
		if (this._actions != null) this._actions = {};
	} else if (fn == null) {
		delete obj[name];
	} else if (name == null) {
		_.each(obj, function (d, n) {
			obj[n] = d.filter(function (f) {
				return f !== fn;
			});
		});
	} else if (obj[name] != null) {
		obj[name] = _.without(obj[name], fn);
	}

	return this;
}

function fire(a, b, args) {
	var action = Action.create(a, b);
	if (!action.context) {
		throw new Error("Action is missing a context.");
	}

	var ctx = action.context;
	var name = action.name;
	args = [].concat(action, args);

	// runs function, unless propagation is stopped
	function run(fn) {
		if (!action.bubbles) return true;
		fn.apply(ctx, args);
	}

	// bubble the action up through all the contexts
	while (action.bubbles && ctx) {
		if (ctx.template) {
			var acts = ctx.template._actions;
			if (acts != null && Array.isArray(acts[name])) {
				acts[name].some(run);
			}
		}

		ctx = ctx.parent;
	}

	// bubble action to the global actions
	if (action.bubbles && Array.isArray(actions[name])) {
		actions[name].some(run);
	}

	return action;
}

var actions$1 = Object.freeze({
	Action: Action,
	plugin: plugin$1,
	default: plugin$1,
	defineEvent: defineEvent,
	add: add$1,
	remove: remove$1,
	fire: fire
});

// import * as AST from "./ast";

var parse$1 = require("./m+xml").parse;
var url = "sourceMappingURL=data:application/json;charset=utf-8;base64,";

function getSource(nodes, src, options) {
	var source = new sourceMap.SourceNode(null, null, options.originalFilename, nodes);
	if (!options.sourceMap) return source.toString();

	if (src) source.setSourceContent(options.originalFilename, src);
	var result = source.toStringWithSourceMap();
	var map64 = new Buffer(result.map.toString(), "utf-8").toString("base64");
	return result.code + "//# " + url + map64;
}

function compile(tree, options) {
	options = _.extend({
		originalFilename: "template.js"
	}, options);

	var src = undefined;
	if (typeof tree === "string") {
		src = tree;
		tree = parse$1(src, options);
	}

	var nodes = _.invoke([].concat(tree), "compile", options);
	return getSource(nodes, src, options);
}

function exec(tpl, options) {
	/* jshint -W054 */
	var r = compile(tpl, options);
	console.log(r);
	return new Function("Temple", r)(require("../"));
}

// export function compileHTML(html, options) {
// 	options = options || {};
// 	var node = new AST.HTML(null, html);
// 	return getSource(node.compile(options), html, options);
// }
//
// export function renderHTML(html, options) {
// 	/* jshint -W054 */
// 	var r = compileHTML(html, options);
// 	// console.log(r);
// 	return (new Function("Temple", r))(require("./"));
// }

function Each(val, vars, ctx, fn, that) {
	return map(val, function (item, key) {
		var nctx = new Context(ctx);

		if (vars.length === 1) nctx.set(vars[0], item);else if (vars.length == 2) {
			nctx.set(vars[0], key);
			nctx.set(vars[1], item);
		}

		return fn.call(that, nctx, key, val);
	});
}

function With(val, ctx, fn, that) {
	var nctx = new Context(ctx);
	nctx.set(val);
	nctx.dataVar.set(val);
	return fn.call(that, nctx);
}

function Element$1(tagname, key, ctx, fn, that) {
	var node = elementOpen(tagname, key);
	var nctx = new Context(null, ctx);
	var comp = autoelement(node, function () {
		return fn.call(that, nctx, node);
	});
	elementClose(tagname);
	return comp;
}

var version = "0.7.0-alpha.2";

var parse = require("./m+xml").parse;
var AST = require("./ast");

exports.version = version;
exports.parse = parse;
exports.AST = AST;
exports.Context = Context;
exports.idom = idom;
exports.utils = utils;
exports.Trackr = Trackr;
exports.Map = trackrObjects.Map;
exports.List = trackrObjects.List;
exports.Variable = trackrObjects.Variable;
exports.plugins = plugins$1;
exports.decorators = decorators$1;
exports.actions = actions$1;
exports.templates = templates;
exports.types = types;
exports.Template = Template;
exports.helpers = helpers;
exports.getHelper = getHelper;
exports.render = render;
exports.getByName = getByName;
exports.paint = paint;
exports.registerType = registerType;
exports.create = create;
exports.getSource = getSource;
exports.compile = compile;
exports.exec = exec;
exports.Each = Each;
exports.With = With;
exports.Element = Element$1;